import { Player, world } from "@minecraft/server";
import { State } from "./State.js";
import { Transition } from "./Transition.js";
import { Vec } from "./Vec.js";

/*

    Clase automata: es el automata como tal. Los jugadores no pueden interactuar con este

*/
export class Automata {

    /**@type {Transition[]} */transitions = [];
    /**@type {State[]} */states = [];
    /**@type {Number[]} */finalStates = [];
    /**@type {Number} */initialState = -1;
    /**@type {String[]} */alphabet = [];
    //other
    name = JSON.stringify(Math.floor(Math.random()*Math.pow(10^15)));
    /**@type {import("@minecraft/server").Vector2} */size = {
        x: 0,
        y: 0
    };


    /**
     * 
     * @param {String} s 
     * @param {State} state 
     * @param {number} i 
     * @returns {{flag: 0|1|2,state:State|undefined,charIndex:Number|undefined}}
     * 
     * Flags:
     * 
     *  0: el estado leyó correctamente la cadena. No envía más información.
     * 
     *  1: La cadena termina en un estado que no pertenece al conjunto de estados finales. Envía el estado.
     * 
     *  2: El último estado en leer la cadena no puede avanzar. Envía el estado y el índice del caracter.
     * 
     */
    readString(s,state=this.states[this.initialState],i=0){
        if(i == s.length)
            return this.finalStates.includes(this.states.indexOf(state)) ?
            { flag: 0 } : { flag: 1, state: state };
        let char = s.charAt(i);
        for(let transition of state.transitions) {
            if(transition.character == char) return this.readString(s,transition.endState,++i);
        }
        return {
            flag: 2,
            state: state,
            charIndex: i
        };
    }

    addChar(char){
        if(!this.alphabet.includes(char)) this.alphabet.push(char);
    }

    removeChar(char){
        this.alphabet.splice(this.alphabet.indexOf(char),1);
    }

    isRedundantChar(char){
        for(let t of this.transitions){
            if(t.character == char) {
                return false;
            }
        }
        return true;
    }

    addState(name,position){
        this.states.push(new State(name,position));
    }

    addTransition(/**@type {Transition} */transition){
        this.transitions.push(transition);
        transition.startState.transitions.push(transition);
    }

    getStateIndex(position){
        for(let i = 0;i<this.states.length;++i){
            if(this.states[i].position.x == position.x && this.states[i].position.y == position.y) return i;
        }
        return -1;
    }

    getState(position){
        return this.states[this.getStateIndex(position)];
    }

    removeState(position){
        let i = this.getStateIndex(position);
        let state = this.states[i];
        this.states.splice(i,1);
        if(this.initialState == i) this.initialState = -1;
        let finalStateI = this.finalStates.indexOf(i);
        if(finalStateI > -1) this.finalStates.splice(finalStateI,1);
        for(let t of state.transitions) this.removeTransition(t,false);
        let transitionsLength = this.transitions.length;
        for(let i = 0; i<transitionsLength;++i) {
            if(this.transitions[i].endState.equals(state)) {
                this.removeTransition(this.transitions[i]);
                --transitionsLength;
                --i;
            } 
        }
        state.isValid = false;
        return state
    }

    removeTransition(/**@type {Transition}*/transition,searchInStates=true){
        transition.isValid = false;
        this.transitions.splice(this.transitions.indexOf(transition),1);
        if(transition.entityName) transition.showChar();
        if(searchInStates) {
            for(let s of this.states){
                let i = s.transitions.indexOf(transition)
                if(i > -1){
                    s.transitions.splice(i,1);
                }
            }
        }
        if(this.isRedundantChar(transition.character)) this.removeChar(transition.character);
    }

    alternFinalState(/**@type {number} */index){
        if(this.finalStates.includes(index)){
            this.finalStates.splice(this.finalStates.indexOf(index),1);
            return false
        } else {
            this.finalStates.push(index);
            return true;
        }
    }

    setInitialState(/**@type {number} */index){
        if(this.initialState != index){
            let lastIndex = this.initialState;
            this.initialState = index;
            return lastIndex;
        }
        return -2;
    }

    /**
     * 
     * @returns {Automata}
     * 
     */
    clone(){
        let automata = new Automata();
        let /**@type {Record<String,Transition[]>} */ finalStatesRecord = {}
        for(let s of this.states){
            let state = new State(s.name,Vec.clone(s.position));
            automata.states.push(state);
            for(let t of s.transitions){
                let transition = new Transition(state,t.character,undefined,[]);
                for(let p of t.unionPositions) transition.unionPositions.push(Vec.clone(p));
                automata.transitions.push(transition);
                state.transitions.push(transition);
                automata.addChar(transition.character)
                if(!finalStatesRecord[t.endState.name]) finalStatesRecord[t.endState.name] = [];
                finalStatesRecord[t.endState.name].push(transition);
            }
        }
        for(let fS of Object.keys(finalStatesRecord)){
            for(let state of automata.states){
                if(state.name == fS){
                    for(let transition of finalStatesRecord[fS])
                        transition.endState = state;
                    break;
                }
            }
        }
        automata.size = Vec.clone(this.size);
        automata.initialState = this.initialState;
        for(let i of this.finalStates) automata.finalStates.push(i);
        return automata;
    }

    /**
     * 
     * @param {String} name 
     * @param {String} author 
     * @param {boolean} private
     * 
     * Guarda el automata. Si tiene un dueño entonces en el acceso público aparecerá el nombre del automata y el nombre del creador,
     * pero si lo marca como privado entonces solo este podrá acceder.
     *  
     */

    getStateType(/**@type {State}*/state){
        let index = this.states.indexOf(state);
        let type = "normal";
        if(this.initialState == index) type = "initial";
        if(this.finalStates.includes(index)) type = type == "initial" ? "fi" : "final";
        return type;
    }

    initialStateToString(){
        return this.initialState > -1 ? this.states[this.initialState].name : "automata.without_initial_state";
    }

    transitionsToString(){
        let transitions = "";
        let i = 0;
        while (i<this.transitions.length){
            transitions+=this.transitions[i].toString();
            if(++i<this.transitions.length) transitions+=", ";
        }
        return transitions;
    }
    
    statesToString(){
        let states = "";
        let i = 0;
        while (i<this.states.length){
            states+=this.states[i].name;
            if(++i<this.states.length) states+=", ";
        }
        return states;
    }
    
    finalStatesToString(){
        let stateF = "";
        let i = 0;
        while (i<this.finalStates.length){
            stateF+=this.states[this.finalStates[i]].name;
            if(++i<this.finalStates.length) stateF+=", ";
        }
        return stateF;
    }

    alphabetToString(){
        let alphabetS = "";
        let i = 0;
        while (i<this.alphabet.length){
            alphabetS+=this.alphabet[i];
            if(++i<this.alphabet.length) alphabetS+=", ";
        }
        return alphabetS;
    }

    toString(){

        return "( "+this.initialStateToString()+
        ", { "+this.alphabetToString()+" }, { "+this.statesToString()+
        " }, { "+this.finalStatesToString()+" }, { "+this.transitionsToString()+" })";
    }

    /**
     * 
     * @param {Automata} automata 
     */
    equals(automata){
        if(this.initialState == automata.initialState && this.finalStates.length == automata.finalStates.length &&
        this.alphabet.length == automata.alphabet.length && this.states.length == automata.states.length &&
        this.transitions.length == automata.transitions.length){
            for(const f of this.finalStates){
                if(!automata.finalStates.includes(f)) return false;
            }
            for(const c of this.alphabet){
                if(!automata.alphabet.includes(c)) return false;
            }
            for(const s of this.states){
                let includes = false;
                for(const sA of automata.states){
                    if(s.equals(sA)) {
                        includes = true;
                        break;
                    }
                }
                if(!includes) return false;
            }
            for(const t of this.transitions){
                let includes = false;
                for(const tA of automata.transitions){
                    if(t.equals(tA)) {
                        includes = true;
                        break;
                    }
                }
                if(!includes) return false;
            }
            return true;
        }
        return false;
    }

    /**
     * 
     * @returns 
     * Devuelve el autómata como un objeto con los nombres de sus atributos acortados y descartando otros como el de las transiciones
     */

    toShortJSON(){
        let statesJ = [];
        for(let s of this.states) statesJ.push(s.toShortJSON());
        return {
            sx: this.size.x,
            sy: this.size.y,
            s: statesJ,
            i: this.initialState,
            f: this.finalStates
        }
    }
    /**
     * 
     * @param {{
     *    sx: number;
     *    sy: number;
     *    s: {
     *        n: string;
     *        t: {
     *            c: string;
     *            sf: string;
     *            p: Vector2[];
     *        }[];
     *        px: number;
     *        py: number;
     *    }[];
     *    i: number;
     *    f: number[];
     * }} data 
     * @returns
     * 
     * Crea un automata a partir de un JSON acortado.
     */
    static getFromShortJSON(data){
        let automata = new Automata();
        let /**@type {Record<String,Transition[]>} */ finalStatesRecord = {}
        for(let s of data.s){
            let state = new State(s.n,{x:s.px,y:s.py});
            automata.states.push(state);
            for(let t of s.t){
                let transition = new Transition(state,t.c,undefined,t.p);
                automata.transitions.push(transition);
                state.transitions.push(transition);
                automata.addChar(transition.character)
                if(!finalStatesRecord[t.sf]) finalStatesRecord[t.sf] = [];
                finalStatesRecord[t.sf].push(transition);
            }
        }
        for(let fS of Object.keys(finalStatesRecord)){
            for(let state of automata.states){
                if(state.name == fS){
                    for(let transition of finalStatesRecord[fS])
                        transition.endState = state;
                    break;
                }
            }
        }
        automata.size.x = data.sx;
        automata.size.y = data.sy;
        automata.initialState = data.i;
        automata.finalStates = data.f;
        return automata;
    }
}