import { Block, BlockPermutation, EquipmentSlot, ItemStack, MolangVariableMap, Player, TicksPerSecond, system, world } from "@minecraft/server";
import { ActionFormData, MessageFormData, ModalFormData } from "@minecraft/server-ui";
import { Automata } from "./Automata.js";
import { State } from "./State.js";
import { Vec } from "./Vec.js";
import { SetTransition } from "./SetTransition.js";
import { Transition } from "./Transition.js";
import { ReadStringTime } from "./ReadStringTime.js";
import { Database } from "./Database/Database.js";
const stateBlockTypes = [
    'minecraft:oak_planks',
    'minecraft:birch_planks',
    'minecraft:spruce_planks',
    'minecraft:dark_oak_planks',
    'minecraft:acacia_planks',
    'minecraft:cherry_planks',
    'minecraft:bamboo_planks',
    "minecraft:crimson_planks",
    "minecraft:warped_planks",
    "minecraft:jungle_planks",
    "minecraft:iron_block",
    "minecraft:gold_block",
    "minecraft:diamond_block",
    "minecraft:emerald_block",
    "minecraft:redstone_block",
    "minecraft:lapis_block",
    "minecraft:netherite_block",
    "minecraft:coal_block",
    'minecraft:stone'
];

for(let c of ["white","orange","magenta","light_blue","yellow","lime","pink","gray","light_gray","cyan","purple","blue","brown","green","red","black"]){
    for(let t of ["wool","concrete","terracotta"]) stateBlockTypes.push("minecraft:"+c+"_"+t);
}


const /**@type {Record<String,PhysicAut>} */playerRecord = {};
const /**@type {Record<String,Transition>} */nameEntityTransitionRecord = {};
const DEFAULT_STATE_NAME_PREFIX = "e";
const DEFAULT_TRANSITION_PARTICLE = "aut:transition_pos";
const TRANSITION_MOLANG_MAP = new MolangVariableMap();
TRANSITION_MOLANG_MAP.setFloat("time",Transition.getUpdateTick()/TicksPerSecond+0.05);
TRANSITION_MOLANG_MAP.setFloat("size",Transition.getPositionSeparation());
TRANSITION_MOLANG_MAP.setFloat("is_final",0);
const DEFAULT_SHOW = true;

/*

    Automata físico: Es un automata con el cual el jugador puede interactuar, haciendo cosas como leer paso a paso, crear/modificar/eliminar estados y transiciones, etc.
    Se puede tomar como el espacio de trabajo

*/
export class PhysicAut extends Automata {
    //physical
    /**@type {import("@minecraft/server").Vector3} */location;
    /**@type {import("@minecraft/server").Dimension} */dimension;
    /**@type {'x'|'z'} */axis;
    /**@type {'x'|'z'} */axisI;
    /**@type {1|-1} */side;
    //actions
    /**@type {{type:'createState'|'removeState'|'modState',data:any}[]} */actions = [];
    /**@type {Number} */actionIndex = -1;
    // show transitions
    transitionParticleType = DEFAULT_TRANSITION_PARTICLE;
    showStates = DEFAULT_SHOW;
    showTransition = DEFAULT_SHOW;
    // state
    stateNamePrefix = DEFAULT_STATE_NAME_PREFIX;
    // runIds
    transitionShowRunId = -1;

    constructor(/**@type {import("@minecraft/server").Dimension} */dimension,
    /**@type {import("@minecraft/server").Vector3} */location,/**@type {'x'|'z'} */axis,/**@type {Number} */sizeX,/**@type {Number} */sizeY,side,id=undefined){
        super();
        this.dimension = dimension;
        this.location = location;
        this.axis = axis;
        this.axisI = axis == 'x' ? 'z' : 'x';
        this.size = {x:sizeX, y:sizeY};
        this.side = side;
        if(this.showTransition) this.transitionShowRunId = system.runInterval(()=>this.showTransitionUpdate(),Transition.getUpdateTick());
        if(id) this.id = id;
    }



    //player interactions
    placeBlock(/**@type {import("@minecraft/server").Block} */block,/**@type {Player} */player){
        if(this.isActionInsideWorkspace(block.location)){
            this.addAction({
                type: "createState",
                data: {
                    state: this.addState(block.location)
                }
            })
        }
    }

    breakBlock(/**@type {Block}*/block,/**@type {BlockPermutation} */p){
        this.addAction({
            type: "removeState",
            data: {
                state: this.removeState(this.getRelativePosition(block.location)),
                blockPermutation: p
            }
        })
    }

    itemUse(/**@type {import("@minecraft/server").ItemUseBeforeEvent} */data){
        data.cancel = true;
        switch(data.itemStack.typeId){
            case 'minecraft:compass':
                system.run(()=>this.openMenu(data.source));
                break;
        }
    }

    itemUseOn(/**@type {import("@minecraft/server").ItemUseOnBeforeEvent} */data){
        if(!stateBlockTypes.includes(data.itemStack.typeId)){
            data.cancel = true;
            if(this.isActionInsideWorkspace(data.block.location)) switch(data.itemStack.typeId){
                case 'aut:altern_final_state':
                    system.run(()=>this.alternFinalState(this.getStateIndex(this.getRelativePosition(data.block.location)),data.source));
                break;
                case 'aut:set_initial_state':
                    system.run(()=>this.setInitialState(this.getStateIndex(this.getRelativePosition(data.block.location)),data.source));
                break;
                case 'aut:set_transition':
                    system.run(()=>this.setTransition(this.getState(this.getRelativePosition(data.block.location)),data.source));
                break;
            } else switch(data.itemStack.typeId){
                case 'aut:set_transition':
                    system.run(()=>this.addTransitionPos(data.block.location,data.source));
                break;
            }
        }
    }

    interactWithName(/**@type {Player} */player,/**@type {import("@minecraft/server").Entity} */entityName){
        let transition = nameEntityTransitionRecord[entityName.id];
        if(transition) {
            let itemStack = player.getComponent("equippable").getEquipment(EquipmentSlot.Mainhand);
            if(itemStack) switch(itemStack.typeId) {
                case "aut:set_transition":
                    this.setTransitionCharUI(transition.startState,transition.endState,player).then((char)=>{
                        if(transition.isValid && char){
                            this.changeChar(transition,char)
                        }
                    })
                    break;
                case "aut:cut_transition":
                    player.playSound("mob.sheep.shear");
                    this.removeTransition(transition);
                    break;
            }
            
        }
    }

    isActionInsideWorkspace(/**@type {import("@minecraft/server").Vector3} */location){
        if(location[this.axisI] - this.location[this.axisI] == this.side){
            let relPos = this.getRelativePosition(location);
            if(relPos.x <= this.size.x && relPos.y <= this.size.y && relPos.x >= 0 && relPos.y >= 0) return true;
        }
        return false;
    }

    //other
    getRelativePosition(/**@type {import("@minecraft/server").Vector3} */location){
        return {
            x: location[this.axis]-this.location[this.axis],
            y: location.y-this.location.y,
        };
    }

    //actions register
    addAction(/**@type {{type:'createState'|'removeState'|'modState',data:any}} */action){
        if(this.actionIndex > -1){
            let newActions = []
            for(let i = 0; i<this.actionIndex+1;++i) newActions.push(this.actions[i])
        }
        this.actions.push(action);
    }

    //actions

    changeChar(transition,char){
        if(this.isRedundantChar(transition.char)) this.removeChar(char);
        transition.setChar(char);
        super.addChar(char);
    }

    addState(/**@type {import("@minecraft/server").Vector3} */location){
        let type = "normal";
        if(this.states.length == 0) {
            this.initialState = 0;
            type = "initial";
        }
        let stateNumber;
        let find;
        for(stateNumber = 0; stateNumber<this.states.length;++stateNumber){
            find = false;
            for(let i = 0;i<this.states.length;++i){
                if(this.states[i].name == this.stateNamePrefix+stateNumber){
                    find = true;
                    break;
                }
            }
            if(!find) break;
        }
        let state = new State(this.stateNamePrefix+stateNumber,this.getRelativePosition(location));
        this.states.push(state);
        if(this.showStates) state.showName(Vec.addPosition(location,0.5,0,0.5),this.dimension,type);
        return state;
    }
    setTransition(/**@type {State} */state,/**@type {Player} */player){
        let st =SetTransition.get(player);
        if(st){
            if(!(state.equals(st.state) && st.positions.length == 0)) {
                if(st.positions.length == 0) {
                    let getPositions = Vec.lineVec2Between(st.state.position,state.position,Transition.getPositionSeparation());
                    for(let i = 0;i<2;++i) getPositions.pop();
                    st.positions = st.positions.concat(getPositions);
                } else {
                    let getPositions = Vec.lineVec2Between(st.positions[st.positions.length-1],state.position,Transition.getPositionSeparation());
                    for(let i = 0;i<2;++i) getPositions.pop();
                    st.positions = st.positions.concat(getPositions);
                }
                this.setTransitionCharUI(st.state,state,player).then((char)=>{
                    if(st.isValid && char){
                        let transition = new Transition(st.state,char,state,st.positions);
                        this.addTransition(transition);
                        this.addChar(char);
                        if(this.showTransition) nameEntityTransitionRecord[
                            transition.showChar(transition.getMidLocation(this.location,this.axis,this.axisI,this.side),this.dimension)] = transition;
                    }
                    st.delete();
                })
            } else player.sendMessage({"translate":"physicAut.setTransition.error.no_start_pos"});

        } else new SetTransition(player,state,this.dimension,this.location,this.axis,this.axisI,this.size,this.side);
    }

    addTransitionPos(/**@type {import("@minecraft/server").Vector3} */loc,/**@type {Player} */player){
        if(loc[this.axisI] == this.location[this.axisI] && loc.y >= this.location.y && loc.y-this.location.y <= this.size.y) {
            let st = SetTransition.get(player);
            if(st) st.addPosition(this.getRelativePosition(loc));
        } 
    }

    setInitialState(/**@type {number} */index,/**@type {Player} */player){
        let result = super.setInitialState(index);
        if(result > -2) {
            player.playSound('bubble.pop');
            if(this.showStates) {
                this.states[index].setEntityName(this.finalStates.includes(index) ? "fi" : "initial");
                if(result > -1)
                    this.states[result].setEntityName(this.finalStates.includes(result) ? "final" : "normal");
            }
        } 
    }
    alternFinalState(/**@type {number} */index,/**@type {Player} */player){
        if(super.alternFinalState(index)){
            if(this.showStates) {
                this.states[index].setEntityName(this.initialState == index ? "fi" : "final")
            } 
            player.playSound("bubble.pop");
        } else {
            if(this.showStates) {
                this.states[index].setEntityName(this.initialState == index ? "initial" : "normal")
            } 
            player.playSound("random.pop2");
        }
        
    }

    removeTransition(transition){
        super.removeTransition(transition);
        ReadStringTime.stopRead(this.id);
    }

    removeState(/**@type {import("@minecraft/server").Vector2} */pos){
        let s = super.removeState(pos);
        s.showName();
        for(let t of s.transitions){
            delete nameEntityTransitionRecord[t.showChar()];
        }
        SetTransition.verifyRemovedState(s);
        ReadStringTime.stopRead(this.id);
        return s;
    }

    //visual

    setTransitionsVisible(b){
        if(this.showTransition != b){
            this.showTransition = b;
            if(this.showTransition) {
                this.showAllTransitionsChar();
                this.transitionShowRunId = system.runInterval(()=>this.showTransitionUpdate(),Transition.getUpdateTick());
            } 
            else {
                ReadStringTime.stopRead(this.id);
                system.clearRun(this.transitionShowRunId);
                for(let t of this.transitions) {
                    delete nameEntityTransitionRecord[t.showChar()]
                } 
            } 
        }
    }

    showAllTransitionsChar(){
        for(let t of this.transitions) {
            if(!t.entityName) {
                nameEntityTransitionRecord[
                    t.showChar(t.getMidLocation(this.location,this.axis,this.axisI,this.side),this.dimension)] = t;
            }
        }
    }

    setStatesVisible(b){
        if(this.showStates != b){
            this.showStates = b;
            if(this.showStates) this.showAllStates();
            else {
                ReadStringTime.stopRead(this.id);
                for(let s of this.states) s.showName();
            }
        }
    }
    showAllStates(){
        for(let s of this.states) {
            if(!s.entityName) s.showName(s.getMidLocation(this.location,this.axis,this.axisI,this.side),this.dimension,this.getStateType(s));
        }
    }
    showTransitionUpdate(){
        let loc = {
            x: 0,
            y: 0,
            z: 0
        }
        loc[this.axisI]+=this.side+0.5;
        for(const t of this.transitions){
            let finalIndex = t.unionPositions.length-1;
            for(let i = 0;i<finalIndex;++i) {
                loc[this.axis] = t.unionPositions[i].x+0.5;
                loc.y = t.unionPositions[i].y+0.5;
                this.dimension.spawnParticle(this.transitionParticleType,Vec.addVec(this.location,loc),TRANSITION_MOLANG_MAP);
            }
            loc[this.axis] = t.unionPositions[finalIndex].x+0.5;
            loc.y = t.unionPositions[finalIndex].y+0.5;
            TRANSITION_MOLANG_MAP.setFloat("is_final",1);
            this.dimension.spawnParticle(this.transitionParticleType,Vec.addVec(this.location,loc),TRANSITION_MOLANG_MAP);
            TRANSITION_MOLANG_MAP.setFloat("is_final",0);
        }
    }

    /**
     * 
     * @param {Player} player
     * Analiza si el bloque en la posición de las transiciones de la hotbar puede ser puesto como muro.
     * En el caso verdadero reemplaza el muro
     *   
     */

    playerChangeWall(player){
        let /**@type {ItemStack} */ itemStack = player.getComponent('inventory').container.getItem(0);
        if(itemStack) {
            if(stateBlockTypes.includes(itemStack.typeId)){
                player.sendMessage({"translate": "physicAut.changeWall.success"});
                this.setWallType(itemStack.typeId);
            } else player.sendMessage({"translate": "physicAut.changeWall.error.invalid_item", with: {"rawtext":[{"translate":`tile.${itemStack.typeId}.name`}]}});
        } else player.sendMessage({"translate": "physicAut.changeWall.error.no_item"});
    }

    setWallType(/**@type {String} */type){
        let finalLocation = Object.assign({},this.location);
        finalLocation[this.axis]+=this.size.x;
        finalLocation.y += this.size.y;
        this.dimension.runCommandAsync(`fill ${this.location.x} ${this.location.y} ${this.location.z} ${finalLocation.x} ${finalLocation.y} ${finalLocation.z} ${type}`);
    }

    clearTransitionBlocks(){
        let loc = Vec.clone(this.location);
        loc[this.axisI]+= this.side;
        let finalLoc = Vec.clone(loc);
        finalLoc[this.axis]+=this.size.x;
        finalLoc.y+=this.size.y;
        this.dimension.runCommandAsync(`fill ${loc.x} ${loc.y} ${loc.z} ${finalLoc.x} ${finalLoc.y} ${finalLoc.z} air`);
    }

    // UIs

    async setTransitionCharUI(/**@type {State} */startState,/**@type {State} */finalState,/**@type {Player} */player,errorMsg=""){
        if(startState.isValid && finalState.isValid){
            const UI = new ModalFormData()
                .title({"translate":"physicAut.setTransitionCharUI.title"})
                .textField({"translate":errorMsg},{translate:"physicAut.setTransitionCharUI.placeHolder"});
            let response = await UI.show(player);
            if(startState.isValid && finalState.isValid && !response.canceled && response.formValues[0].length > 0){
                let /**@type {String} */ char = response.formValues[0].charAt(0);
                for(let t of startState.transitions){
                    if(char == t.character) return this.setTransitionCharUI(startState,finalState,player,"physicAut.setTransitionCharUI.error.same_char");
                }
                return char;
            }
        }
        return undefined;
    }

    async openMenu(/**@type {Player} */player){
        const UI = new ActionFormData()
        .title({translate:"physicAut.menu.title"})
        .button({translate:"physicAut.menu.read_string"})
        .button({translate:"physicAut.menu.get_quintuple"})
        .button({translate:"physicAut.menu.change_wall"})
        .button({translate:"physicAut.menu.save"})
        .button({translate:"physicAut.menu.load"})
        .button({translate:"physicAut.menu.delete_save"});
        const response = await UI.show(player);
        if(!response.canceled) switch (response.selection){
            case 0:
                if(this.initialState > -1) this.readStringMenu(player);
                else player.sendMessage({translate:"automata.readString.no_initial_state"})
                break;
            case 1:
                player.sendMessage(this.toRawMessage());
                break;
            case 2:
                this.playerChangeWall(player);
                break;
            case 3:
                this.saveUI(player);
                break;
            case 4:
                this.loadUI(player);
                break;
            case 5:
                this.deleteSaveUI(player);
                break;
        }
    }

    async readStringMenu(/**@type {Player} */player){
        const UI = new ModalFormData()
        .title({translate:"physicAut.readString.title"})
        .textField({translate:"physicAut.readString.insert_string"},{translate:"physicAut.readString.empty"})
        .dropdown({translate:"physicAut.readString.execution_time"},[
            {translate:"physicAut.readString.instant"},
            {translate:"physicAut.readString.quick"},
            {translate:"physicAut.readString.fast"},
            {translate:"physicAut.readString.normal"},
            {translate:"physicAut.readString.slow"},
            /*{translate:"physicAut.readString.step_by_step"}*/]);
        const response = await UI.show(player);
        if(!response.canceled && this.initialState > -1) {
            ReadStringTime.stopRead(this.id);
            if(response.formValues[1] == 0 || !(this.showStates && this.showTransition)){
                this.readStringMessage(this.readString(response.formValues[0]),response.formValues[0],player)
            }
            /*else if(response.formValues[1] == 5){
                console.warn("pasito a pasito")
            }*/ else {
                new ReadStringTime(this,this.states[this.initialState],response.formValues[0],[5,10,20,30][response.formValues[1]-1],[player]);
            }
        }
    }

    async saveUI(/**@type {Player} */player){
        if(this.initialState > -1){
            if(this.finalStates.length > 0){
                const UI = new ModalFormData()
                .title({translate: "physicAut.save.title"})
                .textField("",{translate: "physicAut.save.placeHolder"})
                //.toggle({translate: "physicAut.save.isPrivate"});
                const response = await UI.show(player);
                if(!response.canceled && this.initialState > -1 && this.finalStates.length > 0){
                    if(Database.isAutomataSaveOverwriting(response.formValues[0],player.name,this,false/*response.formValues[1]*/)){
                        const responseDialog = await this.saveOverwriteWarning(player);
                        if(!response.canceled && responseDialog.selection == 0 && this.initialState > -1 && this.finalStates.length > 0){
                            Database.saveAutomata(response.formValues[0],this,player.name,false/*response.formValues[1]*/);
                            player.sendMessage({translate: "physicAut.save.success"});
                        }
                    } else {
                        Database.saveAutomata(response.formValues[0],this,player.name,response.formValues[1]);
                        player.sendMessage({translate: "physicAut.save.success"});
                    }
                }
            }else player.sendMessage({translate: "physicAut.save.error.no_final_state"})
        } else player.sendMessage({translate: "physicAut.save.error.no_initial_state"})
    }

    async saveOverwriteWarning(/**@type {Player} */player){
        const UI = new MessageFormData()
        .title({translate: "physicAut.save.overwrite.title"})
        .body({translate: "physicAut.save.overwrite.body"})
        .button1({translate: "physicAut.save.overwrite.accept"})
        .button2({translate: "physicAut.save.overwrite.deny"});
        return UI.show(player);
    }

    async loadUI(/**@type {Player} */player){
        let nameList = [];
        for(let n of Database.getPlayerAllSavesNames(player.name)) nameList.push(n);
        if(nameList.length > 0){
            const UI = new ActionFormData()
            .title({translate: "physicAut.load.title"});
            for(let n of nameList) UI.button(n);
            const response = await UI.show(player);
            if(!response.canceled) {
                this.loadAutomata(Database.getAutomata(nameList[response.selection],player.name),
                Database.getPlayerSaveData(player.name).wallType,
                Database.getPlayerSaveData(player.name).stateType);
                this.setStatesVisible(true);
                this.setTransitionsVisible(true);
                player.sendMessage({translate: "physicAut.load.success"});
            }
        } else player.sendMessage({translate: "physicAut.load.error.no_saves"});
       
    }

    async deleteSaveUI(/**@type {Player} */player){
        let nameList = [];
        for(let n of Database.getPlayerAllSavesNames(player.name)) nameList.push(n);
        if(nameList.length > 0){
            const UI = new ActionFormData()
            .title({translate: "physicAut.delete_save.title"});
            for(let n of nameList) UI.button(n);
            const response = await UI.show(player);
            if(!response.canceled) {
                Database.deleteAutomataSave(nameList[response.selection],player.name);
                player.sendMessage({translate: "physicAut.delete_save.success"});
            }
        } else player.sendMessage({translate: "physicAut.delete_save.error.no_saves"});
       
    }

    /**
     * 
     * @param {Automata} automata 
     * @param {string} wallType 
     * @param {string} stateType 
     */
    loadAutomata(automata,wallType,stateType){
        this.clearAll();
        this.setWallType(wallType);
        this.size = automata.size;
        this.alphabet = automata.alphabet;
        this.initialState = automata.initialState;
        this.finalStates = automata.finalStates;
        this.transitions = automata.transitions;
        this.states = automata.states;
        for(let s of this.states){
            let loc = { y: this.location.y+s.position.y};
            loc[this.axis] = this.location[this.axis]+s.position.x;
            loc[this.axisI] = this.location[this.axisI]+this.side;
            this.dimension.runCommandAsync(`setblock ${loc.x} ${loc.y} ${loc.z} ${stateType}`);
        }
    }

    clearAll(){
        this.removeAllStates();
        this.clearTransitionBlocks();
        this.setWallType("minecraft:air");
        this.setTransitionsVisible(false);
        this.setStatesVisible(false);
        this.initialState = -1;
        this.finalStates = [];
        this.alphabet = [];
    }

    removeAllStates(){
        for(const s of this.states) this.removeState(s.position);
    }

    /**
     * 
     * @param {{flag:0|1|2,state:State|undefined,charIndex:Number|undefined}} result 
     * @param {String} string 
     * @param {Player} player
     * 
     * Envía un mensaje al jugador según el resultado de la lectura del string de entrada
     *  
     */

    readStringMessage(result,string,/**@type {Player} */player){
        switch(result.flag){
            case 0:
                player.sendMessage({translate: "automata.readString.success"});
            break;
            case 1:
                player.sendMessage({translate: "automata.readString.is_not_final", with: [result.state.toString()]});
            break;
            case 2:
                let errorString = "§c"+string.substring(0,result.charIndex)+
                "§l>>§d"+
                string.charAt(result.charIndex)+
                "§c<<§r§c"+
                string.substring(result.charIndex+1,string.length);
                player.sendMessage({ rawtext: [
                    { translate: "automata.readString.cant_read", with: [result.state.toString()] },
                    { text: "\n"+errorString+"\n" },
                ]})
            break
        }
    }

    //get & set al registro
    static get(/**@type {Player} */player){
        return playerRecord[player.id];
    }

    static set(/**@type {PhysicAut} */p,/**@type {Player} */player){
        playerRecord[player.id] = p;
    }
    //
    getAutomata(){
        let a = new Automata();
        a.transitions = this.transitions;
        for(let i = 0; i<this.transitions.length;++i){
            a.transitions.push(new Transition(this.transitions[i].startState,this.transitions[i].name,this.transitions[i].endState,this.transitions.position));
        }
        for(let i = 0; i<this.states.length;++i){
            a.states.push(new State(this.states[i].name,this.states[i].position));
        }
        a.alphabet = this.alphabet;
        a.finalStates = this.finalStates;
        a.initialState = this.initialState;
        return a;
    }

    /**
     * 
     * @returns {import("@minecraft/server").RawMessage}
     * 
     * Devuelve la quintupla en su estado RawMessage
     */
    toRawMessage(){
        return {translate:"automata.quintuple",with: {rawtext:[
            {translate: this.initialStateToString()},
            {text: this.alphabetToString()},
            {text: this.statesToString()},
            {text: this.finalStatesToString()},
            {text: this.transitionsToString()}
        ]}}
    }
}